#
#  usage: perl TEcov.pl mapfile.map RM.align -max_mm:[integer] -transposon:[element_name]
#

$mapfile="hsa.merged.nr.seq.map";
$RMout="Hsap.genome.shorttitles_ALU.align";
$transposon="AluY";
$max_mm=0;
$TEcoverage=1;
$TEconserv=1;
$TEstructure=1;
$norm_by_hits=1;
$norm_by_reads=1;

$|=1;
###   parse arguments   ###
unless(-e $ARGV[0])
	{
	print"\nMap file $ARGV[0] not found.\n\n";exit;
	}
unless(-e $ARGV[1])
	{
	print"\nRepeatMasker alignment file $ARGV[1] not found.\n\n";exit;
	}

$mapfile=$ARGV[0];
$RMout=$ARGV[1];

foreach(2..@ARGV-1)
	{
	if($ARGV[$_]=~/^ *-no_coverage *$/)
		{
		$TEcoverage=0;
		}
	elsif($ARGV[$_]=~/^ *-no_conserv *$/)
		{
		$TEconserv=0;
		}
	elsif($ARGV[$_]=~/^ *-no_structure *$/)
		{
		$TEstructure=0;
		}
	elsif($ARGV[$_]=~s/^ *-max_mm://)
		{
		$max_mm=$ARGV[$_];
		$max_mm=~s/ *$//;
		if($max_mm!~/^\d+$/)
			{
			print"\nMaximum number of allowed mismatches must be an integer.\n\n";
			}
		}
	elsif($ARGV[$_]=~s/^ *-transposon://)
		{
		$transposon=$ARGV[$_];
		$transposon=~s/ *$//;
		}
	else
		{
		print"\nDon't know how to treat argument: $ARGV[$_]. I will ignore it.\n\n";
		}
	}
if($TEcoverage+$TEconserv+$TEstructure==0)
	{
	print"\nAll possible analyses deselected.\n\n";exit;
	}
if($TEstructure==1 && !-e"RNAfold.exe")
	{
	print"\nCould not find RNAfold.exe!\nSkip in silico transposon folding.\n\n";
	$TEstructure=0;
	}

###   save reads and hits for each mapped sequences   ###
print"\nRead $mapfile\nSave reads and hits for each mapped sequences...";
%reads=();
%hits=();
open(MAP,$mapfile)||die print"\nCould not open map file $mapfile.\n$!\n\n";
while(<MAP>)
	{
	if($_!~/^trans_id/)
		{
		@d=split("\t",$_);
		next if ($d[5]>$max_mm);
		
		if($norm_by_hits==1)
			{
			$reads{$d[4]}=$d[3];	
			}
		else
			{
			$reads{$d[4]}=1;
			}
		if($norm_by_reads==1)
			{
			$hits{$d[4]}++;	
			}
		else
			{
			$hits{$d[4]}=1;
			}
		}
	}
close MAP;

###   Save sequences of hit loci and calculate hit counts   ###
open(MAP,$mapfile)||die print"\nCould not open map file $mapfile.\n$!\n\n";
%comb=();
%seqs=();
%seqs_rc=();
%lengths=();
while(<MAP>)
	{
	if($_!~/^trans_id/)
		{
		@d=split("\t",$_);
		next if ($d[5]>$max_mm);
		unless($comb{"$d[2] $d[4]}"})
			{
			$comb{"$d[2] $d[4]}"}=1;
			
			$lengths{length$d[2]}=1;
			$seqs{$d[2]}+=$reads{$d[4]}/$hits{$d[4]};
			$d[2]=reverse$d[2];
			$d[2]=~tr/ATGC/TACG/;
			$seqs_rc{$d[2]}+=$reads{$d[4]}/$hits{$d[4]};
			}
		}
	}
undef%comb;
undef%reads;
undef%hits;
print" done.";

###   read RepeatMasker file and map sequences to alignments   ###
open(RM,$RMout)||die print"\nCould not open RepeatMasker alignment file $RMout.\n$!\n\n";
$strand=0;
$cons_start=0;
$target="";
$consensus="";
$index=0;
$t0=time;
%TEhits=();
%structure=();
%conserved=();
%not_conserved=();
while(<RM>)
	{
	$i++;
	if($_=~/^\d/)
		{
		$time=time;
		if($time-$t0>0)
			{
			print" $TEid";
			$t0=time;
			}
		if($_=~/$transposon/i)
			{
			$index=1;
			
			# build hash that links query positions with target positions: key=target position / value=consensus position
			%corr=();
			$pos_a=-1;
			$pos_t=-1;
			$cons_start=$cons_start-$strand;
			$target=~s/ .+\s*//;
			
			%cons_rated=(); # count conservation of a certain position max. twice (insertion in target is evaluated as conserved and not conserved)
			while($pos_a<(length$target)-1)
				{
				$pos_a++;
				if(substr($target,$pos_a,1)ne'-')
					{
					$test_1=substr($target,$pos_a,1);
					$pos_t++;
					}
				if(substr($consensus,$pos_a,1)ne'-')
					{
					$test_2=substr($consensus,$pos_a,1);
					$cons_start+=$strand;
					}
				$corr{$pos_t}=$cons_start;
				
				# optionally collect information on positional conservation of repeat sequence
				if($TEconserv==1&&!$cons_rated{$cons_start})
					{
					if(substr($consensus,$pos_a,1) eq substr($target,$pos_a,1))
						{
						$conserved{$TE}{$cons_start}++;
						}
					elsif(substr($consensus,$pos_a,1)=~/[ATGC]/)
						{
						$not_conserved{$TE}{$cons_start}++;
						}
					$cons_rated{$cons_start}=1;
					
					# save position closest to the 3' end (maximum length)
					if($TElength{$TE}<$cons_start)
						{
						$TElength{$TE}=$cons_start;
						}
					}
				}
			
			# map sequences to target (=sbjct sequence in alignment)
			if($TEcoverage==1)
				{
				foreach$l(keys%lengths)
					{
					$target=~s/-//g;
					foreach$pos(0..(length$target)-$l)
						{
						# check for match in sense (relative to alignment sbjct sequence)
						if($seqs{substr($target,$pos,$l)})
							{
							%rated=();
							foreach$pos_hit($pos..($pos+$l)-1)
								{
								unless($rated{$corr{$pos_hit}})
									{
									$coverage{$TE}{$corr{$pos_hit}}{$strand}+=$seqs{substr($target,$pos,$l)};
									$rated{$corr{$pos_hit}}=1;
									
									# save targeted position closest to the 3' end
									if($max_3p_pos{$TE}<$corr{$pos_hit})
										{
										$max_3p_pos{$TE}=$corr{$pos_hit};
										}
									}
								}
							
							# save hit count and length of seq
							$TEhits{$TE}{$strand}{$l}+=$seqs{substr($target,$pos,$l)};
							}
						
						# check for match in antisense (relative to alignment sbjct sequence)
						if($seqs_rc{substr($target,$pos,$l)})
							{
							%rated=();
							foreach$pos_hit($pos..($pos+$l)-1)
								{
								unless($rated{$corr{$pos_hit}})
									{
									$coverage{$TE}{$corr{$pos_hit}}{$strand*-1}+=$seqs_rc{substr($target,$pos,$l)};
									$rated{$corr{$pos_hit}}=1;
									
									# save targeted position closest to the 3' end
									if($max_3p_pos{$TE}<$corr{$pos_hit})
										{
										$max_3p_pos{$TE}=$corr{$pos_hit};
										}
									}
								}
							
							# save hit count and length of seq
							$TEhits{$TE}{$strand*-1}{$l}+=$seqs_rc{substr($target,$pos,$l)};
							}
						}
					}
				}
			
			# fold repeat transcript with RNAfold and save paired/unpared positions
			if($TEstructure==1)
				{
				open(RNAFOLD_IN,">rnafold_in.$transposon.temp");
				print RNAFOLD_IN">$TE\n$target";
				close RNAFOLD_IN;
				system("RNAfold.exe --noLP --noPS < rnafold_in.$transposon.temp > rnafold_out.$transposon.temp");
				open(RNAFOLD_OUT,"rnafold_out.$transposon.temp");
				@rnafold_out=<RNAFOLD_OUT>;
				close RNAFOLD_OUT;
				$rnafold_out[2]=~s/\s*$//;
				$rnafold_out[2]=~s/ .+$//;
				@structure=split('',$rnafold_out[2]);
				foreach$pos(0..@structure-1)
					{
					if($structure[$pos]=~/[\(\)]/)
						{
						$structure{$TE}{'paired'}{$corr{$pos}}++;
						}
					else
						{
						$structure{$TE}{'unpaired'}{$corr{$pos}}++;
						}
					}
				}
			
			# extract information from head
			@head=split(' ',$_);
			if($head[8]eq'C')
				{
				$strand=-1;
				$TE=$head[9];
				$TEid=$head[14];
				}
			else
				{
				$strand=1;
				$TE=$head[8];
				$TEid=$head[13];
				}
			# reset variables
			$i=0;
			$cons_start=0;
			$target="";
			$consensus="";
			}
		else
			{
			$index=0;
			}
		}
	elsif($index==1)
		{
		if($i==2)
			{
			$target.=substr($_,27,50);
			}
		elsif($i==4)
			{
			$i=0;
			$consensus.=substr($_,27,50);
			if($cons_start==0)
				{
				$_=~s/^.{15} +//;
				$_=~s/\d+//;
				$cons_start=$&;
				}
			}
		}
	}
close RM;

###   unlink temporary files from RNAfold   ###
if($TEstructure==1)
	{
	unlink"rnafold_in.$transposon.temp";
	unlink"rnafold_out.$transposon.temp";
	}

if($transposon=~/^\s*$/)
	{
	$transposon='all';
	}

###   print coverage results to output   ###
if($TEcoverage==1)
	{
	open(COVERAGE,">coverage-$transposon.table");
	foreach$TE(keys%coverage)
		{
		foreach$pos(1..$max_3p_pos{$TE})
			{
			unless($coverage{$TE}{$pos}{'1'})
				{
				$coverage{$TE}{$pos}{'1'}=0;
				}
			unless($coverage{$TE}{$pos}{'-1'})
				{
				$coverage{$TE}{$pos}{'-1'}=0;
				}
			print COVERAGE"$TE\t$pos\t$coverage{$TE}{$pos}{'1'}\t-$coverage{$TE}{$pos}{'-1'}\n";
			}
		print COVERAGE"\n";
		}
	close COVERAGE;
	}

###   print conservation results to output   ###
if($TEconserv==1)
	{
	open(COSERVATION,">conservation-$transposon.table");
	foreach$TE(keys%TElength)
		{
		foreach$pos(1..$TElength{$TE})
			{
			unless($conserved{$TE}{$pos})
				{
				$conserved{$TE}{$pos}=0;
				}
			unless($not_conserved{$TE}{$pos})
				{
				$not_conserved{$TE}{$pos}=0;
				}
			$percent_id="n.a.";
			if($conserved{$TE}{$pos}+$not_conserved{$TE}{$pos}>0)
				{
				$percent_id=($conserved{$TE}{$pos}/($conserved{$TE}{$pos}+$not_conserved{$TE}{$pos}))*100;	
				}
			print COSERVATION"$TE\t$pos\t$conserved{$TE}{$pos}\t$not_conserved{$TE}{$pos}\t$percent_id\n";
			}
		print CONSERVATION"\n";
		}
	close CONSERVATION;
	}

###   print TE structure results to output   ###
if($TEstructure==1)
	{
	open(STRUCTURE,">structure-$transposon.table");
	foreach$TE(keys%TEhits)
		{
		foreach$pos(1..$TElength{$TE})
			{
			$percent_paired="n.a.";
			unless($structure{$TE}{'paired'}{$pos})
				{
				$structure{$TE}{'paired'}{$pos}=0;
				}
			unless($structure{$TE}{'unpaired'}{$pos})
				{
				$structure{$TE}{'unpaired'}{$pos}=0;
				}
			if($structure{$TE}{'unpaired'}{$pos}>0)
				{
				$percent_paired=($structure{$TE}{'paired'}{$pos}/($structure{$TE}{'paired'}{$pos}+$structure{$TE}{'unpaired'}{$pos}))*100;
				}
			
			print STRUCTURE"$TE\t$pos\t$structure{$TE}{'paired'}{$pos}\t$structure{$TE}{'unpaired'}{$pos}\t$percent_paired\n";
			}
		print STRUCTURE"\n";
		}
	}

###   print  hit counts per TE to output   ###
if($TEcoverage==1)
	{
	open(HITCOUNTS,">hitcounts-$transposon.table");
	@lengths=keys%lengths;
	@lengths=sort{$a<=>$b}@lengths;
	$TEhits{$TE}{$strand*-1}{$l}+=$seqs_rc{substr($target,$pos,$l)};
	foreach$TE(keys%TEhits)
		{
		foreach$l(@lengths)
			{
			$sum=$TEhits{$TE}{'1'}{$l}+$TEhits{$TE}{'-1'}{$l};
			print HITCOUNTS"$TE\t$l\t$TEhits{$TE}{'1'}{$l}\t$TEhits{$TE}{'-1'}{$l}\t$sum\n";
			}
		print HITCOUNTS"\n";
		}
	close HITCOUNTS;
	}
